# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4
# GPL 2013

import json
import mimetypes
import os
import shutil
import urlparse

import ox

from torrent import TorrentClient
from utils import get_sha1, makedirs, get_user_folder
from version import __version__, __appname__

class API(ox.API):
    __version__ = __version__
    __name__ = __appname__

class Backend:
    sites = {}
    def __init__(self, config):
        self.config = config
        options = None
        if 'port' in config:
            options = {}
            options['port'] = int(config['port']) + 10
        self.t = TorrentClient(options)

    def get(self, key, default=None):
        return self.config.get(key, default)

    def site(self, site, connect=True):
        site = urlparse.urlparse(site).netloc
        if connect and not site in self.sites:
            self.sites[site] = Site(self.t, site)
        return self.sites.get(site)

class Site:
    connected = False

    def __init__(self, t, name):
        self.t = t
        self.name = name
        self.site_config = os.path.join(get_user_folder(), 'config', '%s.json' % self.name)
        if os.path.exists(self.site_config):
            with open(self.site_config) as fd:
                self.config = json.load(fd)
        else:
            self.config = {
                'url': 'http://%s/api/' % name,
                'path': os.path.join(get_user_folder(), 'data', self.name),
                'files': []
            }
        self.connect()

    def add_requests(self, ids):
        r = self.api.addRequests(ids=ids)
        if 'data' in r and 'metadata' in r['data']:
            for item in r['data']['metadata']:
                self.download(item['sha1'], item['info_hash'])

    def connect(self):
        self.connected = False
        if self.config.get('url'):
            self.api = API(self.config['url'])
            if 'username' in self.config:
                r = self.api.signin(username=self.config['username'], password=self.config['password'])
                self.connected = r.get('status', {}).get('code') == 200

    def download(self, sha1, info_hash):
        def done(s):
            del self.t.downloads[sha1]
            data_path = os.path.splitext(self.t.torrent_path(sha1))[0]
            target = os.path.join(self.config['path'], 'new', '%s.mp3' % sha1)
            makedirs(os.path.dirname(target))
            shutil.move(data_path, target)
            os.symlink(target, data_path)
            r = self.api.sync(add=[sha1])
            self.config['files'].append(info(target))
            self.save_config()

        if sha1 not in self.t.handles:
            self.t.download(sha1, info_hash, done)

    def get_file(self, sha1):
        for f in self.config['files']:
            if f['sha1'] == sha1:
                return f
        return None

    def save_config(self):
        makedirs(os.path.dirname(self.site_config))
        with open(self.site_config, 'w') as fd:
            json.dump(self.config, fd, indent=1)

    def seed_requests(self):
        r = self.api.getRequests()
        if 'data' in r and 'ids' in r['data']:
            requested = r['data']['ids']
            for sha1 in requested:
                if sha1 not in self.t.handles:
                    f = self.get_file(sha1)
                    if f:
                        self.t.seed(f['path'], f['sha1'])
            for sha1 in self.t.handles.keys():
                if sha1 not in self.t.downloads and sha1 not in requested:
                    self.t.remove(sha1)

    def sync(self):
        # dont scan if path does not exist at all,
        # might be on external drive
        if not os.path.exists(self.config['path']):
            return
        current = {f['path']: f for f in self.config['files']}
        new = scan(self.config['path'], current, self.config.get('extensions', []))
        for f in new:
            current[f] = new[f]
        removed = []
        for f in current:
            if not os.path.exists(f):
                removed.append(current[f]['sha1'])
                del current[f]
        self.config['files'] = current.values()
        for f in self.config['files']:
            if not 'info_hash' in f:
                tinfo = self.t.make_torrent(f['path'], f['sha1'])
                f['info_hash'] = tinfo.info_hash().to_string().encode('hex')

        r = self.api.sync()
        if 'data' in r and 'ids' in r['data']:
            known = set(r['data']['ids'])
            files = list(set([f['sha1'] for f in current.values()]))
            removed = list(known - set(files))
            added = list(set([f['sha1'] for f in current.values() if f['sha1'] not in known]))
            r = self.api.sync({'add': added, 'remove': removed})
            if 'data' in r and 'update' in r['data']:
                update = []
                by_sha1 = {f['sha1']: f for f in current.values()}
                for sha1 in r['data']['update']:
                    f = by_sha1[sha1]
                    i = {}
                    for key in ('sha1', 'extension', 'info_hash', 'size'):
                        i[key] = f[key]
                    i['title'] = os.path.basename(f['path'])
                    update.append(i)
                if update:
                    r = self.api.sync({'update': update})
            self.save_config()


def info(path):
    info = {}
    stat = os.stat(path)
    info['path'] = path
    info['size'] = stat.st_size
    info['mtime'] = stat.st_mtime
    info['extension'] = os.path.splitext(path)[-1][1:]
    info['type'] = mimetypes.guess_type(path)[0]
    info['sha1'] = get_sha1(path)
    return info

def scan(path, current, extensions):
    if isinstance(path, str):
        path = path.decode('utf-8')
    changed = {}
    for root, folders, files in os.walk(path, followlinks=True):
        for f in files:
            f = os.path.join(root, f)
            ext = os.path.splitext(f)[-1][1:]
            if not extensions or ext in extensions:
                if f not in current:
                    changed[f] = info(f)
                else:
                    stat = os.stat(f)
                    i = current[f]
                    if stat.st_size != i['size'] or int(stat.st_mtime) != int(i['mtime']):
                        changed[f] = info(f)
    return changed
