# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4
# GPL 2013

import inspect
import json
import os
from urlparse import urlparse

from twisted.web.resource import Resource
from twisted.web.static import File

from version import __appname__, __version__
from utils import trim


def json_response(data=None, status=200, text='ok'):
    if not data:
        data = {}
    return {'status': {'code': status, 'text': text}, 'data': data}

class ApiActions(dict):
    properties = {}
    def __init__(self):

        def api(backend, site, data):
            '''
                returns list of all known api actions
                param data {
                    docs: bool
                }
                if docs is true, action properties contain docstrings
                return {
                    status: {'code': int, 'text': string},
                    data: {
                        actions: {
                            'api': {
                                cache: true,
                                doc: 'recursion'
                            },
                            'hello': {
                                cache: true,
                                ..
                            }
                            ...
                        }
                    }
                }
            '''
            docs = data.get('docs', False)
            code = data.get('code', False)
            _actions = self.keys()
            _actions.sort()
            actions = {}
            for a in _actions:
                actions[a] = self.properties[a]
                if docs:
                    actions[a]['doc'] = self.doc(a)
                if code:
                    actions[a]['code'] = self.code(a)
            response = json_response({'actions': actions})
            return response
        self.register(api)

    def doc(self, f):
        return trim(self[f].__doc__)

    def code(self, name):
        f = self[name]
        if name != 'api' and hasattr(f, 'func_closure') and f.func_closure:
            f = f.func_closure[0].cell_contents
        info = f.func_code.co_firstlineno
        return info, trim(inspect.getsource(f))

    def register(self, method, action=None, cache=True):
        if not action:
            action = method.func_name
        self[action] = method
        self.properties[action] = {'cache': cache}

    def unregister(self, action):
        if action in self:
            del self[action]

    def render(self, backend, site, action, data):
        if action in self:
            result = self[action](backend, site, data)
        else:
            result = json_response(status=404, text='not found')
        return json.dumps(result)

actions = ApiActions()


class Server(Resource):

    def __init__(self, base, backend):
        self.base = base
        self.backend = backend
        Resource.__init__(self)

    def static_path(self, path):
        return os.path.join(self.base, 'static', path)

    def get_site(self, request):
        headers = request.getAllHeaders()
        if 'origin' in headers:
            request.headers['Access-Control-Allow-Origin'] = headers['origin']
            site = headers['origin']
        elif 'referer' in headers:
            u = urlparse(headers['referer'])
            site = u.scheme + '://' + u.netloc
        else:
            site = None
        return site

    def getChild(self, name, request):
        request.headers['Server'] = '%s/%s' % (__appname__, __version__)
        site = self.get_site(request)
        if name in ('icon.png', 'favicon.ico'):
            f = File(self.static_path('icon.png'))
            f.isLeaf = True
            return f
        if request.path == '/api/':
            return self
        if request.path.startswith('/get/'):
            sha1 = request.path.split('/')[-1].lower()
            if site:
                site = self.backend.site(site, False)
            if site:
                f = site.get_file(sha1)
                if f:
                    request.headers['Access-Control-Allow-Origin'] = '*'
                    f = File(f['path'], 'audio/mp3')
                    f.isLeaf = True
                    return f
        path = request.path
        path = path[1:]
        if not path:
            path = 'index.html'
        path = self.static_path(path)
        f = File(path)
        if not os.path.isdir(path):
            f.isLeaf = True
        return f

    def render_OPTIONS(self, request):
        # allow all requested headers to sites
        headers = request.getAllHeaders()
        if 'access-control-request-headers' in headers:
            request.headers['Access-Control-Allow-Headers'] = headers['access-control-request-headers']
        return ''

    def render_POST(self, request):
        site = self.get_site(request)
        if 'action' in request.args:
            if 'data' in request.args:
                data = json.loads(request.args['data'][0])
            else:
                data = {}
            action = request.args['action'][0]
            return actions.render(self.backend, site, action, data)

    def render_GET(self, request):
        path = self.static_path('index.html')
        with open(path) as f:
            data = f.read()
        request.headers['Content-Type'] = 'text/html'
        return data
