# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4
# GPL 2013

import datetime
import os
import time
from threading import Thread
from urllib import quote

import libtorrent as lt

from utils import ishex, get_sha1, makedirs, get_user_folder
from version import __appname__, __version__

TRACKER_URL = 'udp://tracker.openbittorrent.com:80/announce'


class TorrentClient:
    path = os.path.join(get_user_folder(), 'torrents')
    progress = {}
    handles = {}
    downloads = {}

    def __init__(self, options=None):
        settings = lt.session_settings()
        settings.user_agent = '%s/%s' % (__appname__, __version__)

        self.session = lt.session()
        if options:
            if options.get('path'):
                self.path = options['path']
            if options.get('max_download_rate'):
                self.session.set_download_rate_limit(int(options['max_download_rate']))
            if options.get('max_upload_rate'):
                self.session.set_upload_rate_limit(int(options['max_upload_rate']))
            if options.get('port'):
                self.session.listen_on(options['port'], options['port'] + 10)
        else:
            self.session.listen_on(6881, 6891)
        self.session.set_settings(settings)
        if options and options.get('proxy_host'):
            ps = lt.proxy_settings()
            ps.type = lt.proxy_type.http
            ps.hostname = options['proxy_host'].split(':')[0]
            ps.port = int(options['proxy_host'].split(':')[1])
            self.session.set_proxy(ps)

        makedirs(self.path)

    def add_magnet(self, sha1, info_hash):
        torrent_path = self.torrent_path(sha1)
        save_path = os.path.dirname(torrent_path)
        atp = {}
        atp['url'] = 'magnet:?xt=urn:btih:%s&dn=%s&tr=%s' % (info_hash, sha1, quote(TRACKER_URL))
        if isinstance(atp['url'], unicode):
            atp['url'] = atp['url'].encode('utf-8')
        atp["save_path"] = save_path
        atp["storage_mode"] = lt.storage_mode_t.storage_mode_sparse
        atp["paused"] = False
        atp["auto_managed"] = True
        atp["duplicate_is_error"] = True
        h = self.session.add_torrent(atp)
        h.set_max_connections(60)
        h.set_max_uploads(-1)
        self.handles[sha1] = h
        return h

    def add_torrent(self, torrent_path):
        save_path = os.path.dirname(torrent_path)
        info = lt.torrent_info(torrent_path)
        sha1 = info.name()
        if len(sha1) == 40 and ishex(sha1):
            if not list(info.trackers()):
                info.add_tracker(TRACKER_URL, 0)
            atp = {}
            atp["save_path"] = save_path
            atp["ti"] = info
            fastresume = torrent_path + '.fastresume'
            if os.path.exists(fastresume):
                try:
                    with open(fastresume, 'rb') as fd:
                        atp["resume_data"] = fd.read()
                    os.unlink(fastresume)
                except:
                    pass
            atp["storage_mode"] = lt.storage_mode_t.storage_mode_sparse
            atp["paused"] = False
            atp["auto_managed"] = True
            atp["duplicate_is_error"] = True
            h = self.session.add_torrent(atp)
            h.set_max_connections(60)
            h.set_max_uploads(-1)
            self.handles[sha1] = h
            #merkel based torrents dont seed otherwise? still not working
            #h.force_recheck()
            return h
        else:
            raise IOError, 'torrent name does not match contents'

    def download(self, sha1, info_hash, callback=None):
        torrent = self.torrent_path(sha1)
        if not os.path.exists(torrent):
            makedirs(os.path.dirname(torrent))
            self.downloads[sha1] = {}
            self.downloads[sha1]['done'] = False
            self.add_magnet(sha1, info_hash)
            t = Thread(target=self.download_update, args=(sha1, callback))
            t.daemon = True
            t.start()
            return True
        return False

    def download_update(self, sha1, callback=None):
        h = self.handles[sha1]
        while True:
            info = self.info(sha1)
            if info.get('total_wanted'):
                self.downloads[sha1]['progress'] = float(info['total_download']) / info['total_wanted']
            if h.is_seed():
                break
            time.sleep(1)
        self.remove(sha1)
        self.downloads[sha1]['progress'] = 1
        self.downloads[sha1]['done'] = True
        if callback:
            callback(sha1)

    def info(self, sha1=None):
        info = {}
        if not sha1:
            return [self.info(sha1) for sha1 in self.handles]

        h = self.handles.get(sha1)
        if h:
            status = h.status()
            for key in dir(status):
                if not key.startswith('__') and key not in ('states', ):
                    info[key] = getattr(status, key)
                    if isinstance(info[key], lt.big_number):
                        info[key] = info[key].to_string().encode('hex')
                    elif isinstance(info[key], status.states):
                        if key == 'state':
                            info[key] = str(info[key])
                        else:
                            del info[key]
                    elif isinstance(info[key], lt.storage_mode_t):
                        info[key] = str(info[key])
                    elif isinstance(info[key], datetime.timedelta):
                        info[key] = info[key].total_seconds()
            info['state'] = [
                'queued', 'checking', 'downloading metadata',
                'downloading', 'finished', 'seeding', 'allocating',
                'checking fastresume'
            ][status.state]
            info['sha1'] = sha1
            info['name'] = h.name()
            #fails with magent links
            #info['file_progress'] = h.file_progress()
            info['peers'] = []
            for p in h.get_peer_info():
                pinfo = {}
                for key in dir(p):
                    if not key.startswith('__'):
                        pinfo[key] = getattr(p, key)
                        if isinstance(pinfo[key], lt.big_number):
                            pinfo[key] = pinfo[key].to_string().encode('hex')

                pinfo['flags'] = []
                for flag in dir(lt.peer_info):
                    if not flag.startswith('__') \
                            and isinstance(getattr(lt.peer_info, flag), int) \
                            and p.flags & getattr(lt.peer_info, flag):
                        pinfo['flags'].append(flag)
                info['peers'].append(pinfo)
        return info

    def make_torrent(self, filename, sha1=None):
        filename = os.path.abspath(filename)
        if not sha1:
            sha1 = get_sha1(filename)
        torrent_path = self.torrent_path(sha1)

        fs = lt.file_storage()
        lt.add_files(fs, filename)

        #with merkel path hash, not seeding right now
        #t = lt.create_torrent(fs, 0, 4 * 1024 * 1024, 2)
        t = lt.create_torrent(fs, 0, 4 * 1024 * 1024)

        t.add_tracker(TRACKER_URL)
        #t.set_creator('%s %s' % (__name, __version__))
        self.progress[sha1] = 0

        def progress(x):
            self.progress[sha1] += 1

        lt.set_piece_hashes(t, os.path.split(filename)[0], progress)
        torrent = t.generate()
        torrent['info']['name'] = sha1
        makedirs(os.path.dirname(torrent_path))
        with open(torrent_path, 'w') as fd:
            fd.write(lt.bencode(torrent))
        data = os.path.splitext(torrent_path)[0]
        if os.path.exists(data):
            os.unlink(data)
        os.symlink(filename, data)
        return lt.torrent_info(torrent_path)

    def remove(self, sha1):
        h = self.handles.get(sha1)
        if h:
            self.session.remove_torrent(h)
            del self.handles[sha1]
            return True
        return False

    def seed(self, filename, sha1=None):
        filename = os.path.abspath(filename)
        if not sha1:
            sha1 = get_sha1(filename)
        if not sha1 in self.handles:
            torrent = self.torrent_path(sha1)
            if not os.path.exists(torrent):
                self.make_torrent(filename)
            return self.add_torrent(torrent)

    def shutdown(self):
        for h in self.handles.values():
            if not h.is_valid() or not h.has_metadata():
                continue
            data = lt.bencode(h.write_resume_data())
            with open(self.torrent_path(h.name()) + '.fastresume', 'wb') as fd:
                fd.write(data)

    def torrent_path(self, sha1):
        #return os.path.join(self.path, sha1[:2], sha1[2:4], sha1[4:6], '%s.torrent' % sha1)
        return os.path.join(self.path, '%s.torrent' % sha1)

